import os
import sys
import click
# noinspection PyUnresolvedReferences
from wrktoolbox import stores, version
from wrktoolbox.benchmarks import BenchmarkSuite
# noinspection PyUnresolvedReferences
from wrktoolbox.goals import *
from wrktoolbox.logs import get_app_logger
from wrktoolbox.commands import get_configuration, SettingsFileNotFound
from rocore.exceptions import InvalidArgument
from roconfiguration import ConfigurationError


logger = get_app_logger()


def run_core(settings):
    sys.path.insert(0, '.')

    try:
        configuration = get_configuration(settings)
    except SettingsFileNotFound as e:
        logger.info(f'[*] Error: {e}')
        exit(1)
    except (InvalidArgument, ConfigurationError):
        logger.exception('An error occurred while loading configuration')
        exit(2)
        return

    logger.info(f'Using settings file {settings}')

    try:
        suite = BenchmarkSuite.from_dict(configuration.values)
    except Exception:
        logger.exception('An error occurred while preparing the suite of benchmarks')
        exit(1)
        return

    suite.configuration = configuration

    if 'metadata' in configuration:
        suite.metadata = configuration.metadata.values

    if not suite.configurations:
        logger.error('Missing benchmark configurations, exiting')
        exit(1)
        return

    if suite.plugins:
        logger.info('Loaded plugins:')
        for plugin in suite.plugins:
            logger.info(f' - {plugin.name}')

        logger.info('---')

    if not suite.goals and not any(configuration.goals for configuration in suite.configurations):
        logger.info('No performance goals are defined for this suite of benchmarks')

    if any(plugin.has_setup for plugin in suite.plugins):
        logger.info('Running setup functions for plugins:')
        for plugin in suite.plugins:
            if plugin.has_setup:
                logger.info(f' - {plugin.name}')
                plugin.setup(suite, logger)

        logger.info('---')

    if 'scripts_folder' in configuration.values:
        scripts_folder = configuration.scripts_folder
        if scripts_folder:
            if not os.path.exists(scripts_folder):
                logger.error('Invalid scripts_folder: path not found')
                exit(2)
            if not os.path.isdir(scripts_folder):
                logger.error('Invalid scripts folder: ')
                exit(2)

    suite.run(logger)

    logger.info('Suite completed successfully')


@click.command(name='run')
@click.option('--settings',
              default='settings.yaml',
              help='Settings source (YAML or JSON); can be a file path or an URL.',
              show_default=True)
def run_command(settings):
    try:
        run_core(settings)
    except KeyboardInterrupt:
        logger.info('[*] User interrupted')
        exit(1)



from abc import abstractmethod
from typing import Sequence, Generator
from rocore.registry import Registry
from wrktoolbox.benchmarks import BenchmarkSuite, BenchmarkOutput


class SuiteReport:
    """A set of results, consisting of the benchmark suite and its results."""

    def __init__(self,
                 suite: BenchmarkSuite,
                 results: Sequence[BenchmarkOutput] = None):
        self.suite = suite
        self.results = results


class ResultsImporter(Registry):
    """A class that can import benchmark results, stored by an implementation of BenchmarkOutputStore"""

    @abstractmethod
    def import_suites(self) -> Generator[SuiteReport, None, None]:
        """Imports all suites stored in the handled source."""

    @abstractmethod
    def import_results(self, report: SuiteReport) -> Generator[BenchmarkOutput, None, None]:
        """Imports the results of a suite."""




import logging
import logging.handlers
from datetime import datetime
from rocore.folders import ensure_folder


logger = None


def get_app_logger():
    global logger

    if logger is not None:
        return logger

    logger = logging.getLogger('wrktoolbox')

    logger.setLevel(logging.INFO)

    max_bytes = 24 * 1024 * 1024

    file_handler = logging.handlers.RotatingFileHandler

    now = datetime.now()
    ts = now.strftime('%Y%m%d')

    ensure_folder(f'logs/{ts}')

    handler = file_handler(f'logs/{ts}/wrktoolbox.log',
                           maxBytes=max_bytes,
                           backupCount=5)

    handler.setLevel(logging.DEBUG)

    logger.addHandler(handler)
    logger.addHandler(logging.StreamHandler())

    return logger



from typing import Optional, Sequence, List
from logging import Logger
from rocore.models import Model, Collection, Boolean
from rocore.exceptions import InvalidArgument
from wrktoolbox.benchmarks import BenchmarkPlugin, handle_plugins, BenchmarkOutput
from wrktoolbox.results import ResultsImporter, SuiteReport
from wrktoolbox.reports import ReportWriter
# noinspection PyUnresolvedReferences
from wrktoolbox.results.importers.fs import JsonResultsImporter, BinResultsImporter


class ReportGeneration(Model):
    importers = Collection(ResultsImporter, nullable=False)
    writers = Collection(ReportWriter, nullable=False)
    plugins = Collection((str, BenchmarkPlugin))
    sort = Boolean()

    def __init__(self,
                 importers: Sequence[ResultsImporter],
                 writers: Sequence[ReportWriter],
                 plugins: Optional[Sequence[BenchmarkPlugin]] = None,
                 sort: bool = True):
        if sort is None:
            sort = True
        self.importers = importers
        self.writers = writers
        self.plugins = plugins
        self.sort = sort

    def _get_reports(self, importer):
        if self.sort:
            items = list(importer.import_suites())
            items.sort(key=lambda item: item.suite.location)

            yield from items
        else:
            yield from importer.import_suites()

    def _get_results(self, importer, report):
        if self.sort:
            items = list(importer.import_results(report))  # type: List[BenchmarkOutput]
            items.sort(key=lambda item: item.url)

            yield from items
        else:
            yield from importer.import_results(report)

    def run(self, logger: Logger):
        if not self.importers:
            raise InvalidArgument('no configured importers')

        if not self.writers:
            raise InvalidArgument('no configured writers')

        for importer in self.importers:

            for report in self._get_reports(importer):  # type: SuiteReport
                logger.info('Imported suite %s', report.suite.id)

                for writer in self.writers:
                    writer.write(report)

                for result in self._get_results(importer, report):  # type: BenchmarkOutput
                    for writer in self.writers:
                        writer.write_output(report, result)

        for writer in self.writers:
            if hasattr(writer, 'close'):
                logger.info('Closing writer %s', writer.get_class_name())
                writer.close()

        logger.debug('Finished processing report')

    @classmethod
    def from_dict(cls, data):
        plugins = list(handle_plugins(data))

        return cls(
            [ResultsImporter.from_configuration(item) for item in data.get('importers')],
            [ReportWriter.from_configuration(item) for item in data.get('writers')],
            plugins
        )



import sys
import click
from wrktoolbox.logs import get_app_logger
from roconfiguration import ConfigurationError
from wrktoolbox.commands import get_configuration, SettingsFileNotFound
from rocore.exceptions import InvalidArgument
# noinspection PyUnresolvedReferences
from wrktoolbox.goals import *
# noinspection PyUnresolvedReferences
from wrktoolbox.stores import *
from wrktoolbox.reports.generation import ReportGeneration


logger = get_app_logger()


def reports_core(settings):
    sys.path.insert(0, '.')

    try:
        configuration = get_configuration(settings)
    except SettingsFileNotFound as e:
        logger.info(f'[*] Error: {e}')
        exit(1)
    except (InvalidArgument, ConfigurationError):
        logger.exception('An error occurred while loading configuration')
        exit(2)
        return

    logger.info(f'Using settings file {settings}')

    # NB: first use an importer to import

    try:
        generation = ReportGeneration.from_dict(configuration.values)
    except Exception:
        logger.exception('An error occurred while preparing the suite of benchmarks')
        exit(1)
        return

    if generation.plugins:
        logger.info('Loaded plugins:')
        for plugin in generation.plugins:
            logger.info(f' - {plugin.name}')

        logger.info('---')

    generation.run(logger)


@click.command(name='reports')
@click.option('--settings',
              default='reports.yaml',
              help='Settings source (YAML or JSON); can be a file path or an URL.',
              show_default=True)
def reports_command(settings):
    try:
        reports_core(settings)
    except KeyboardInterrupt:
        logger.info('[*] User interrupted')
        exit(1)




from typing import Union
from wrktoolbox.benchmarks import PerformanceGoal
from wrktoolbox.wrkoutput import BenchmarkOutput


LimitType = Union[int, float, str]


class NoSocketErrorsGoal(PerformanceGoal):
    """A goal that is satisfied if there are no socket errors.
    Note that timeout errors can easily occur if wrk timeout is set to a small value."""

    type_name = 'no-socket-errors'

    def is_satisfied(self, output: BenchmarkOutput) -> bool:
        return output.socket_errors is None

    def __repr__(self):
        return 'No socket errors'


class NoFailedRequestsGoal(PerformanceGoal):
    """A goal that is satisfied if there are no responses with non-2xx and non-3xx status"""

    type_name = 'no-failed-requests'

    def is_satisfied(self, output: BenchmarkOutput) -> bool:
        return not output.not_successful_responses

    def __repr__(self):
        return 'No responses with status not in 2xx or 3xx'


class NoErrorsGoal(PerformanceGoal):
    """A goal that is satisfied if there are no errors:
    no socket errors and no responses with status not in 2xx or 3xx."""

    type_name = 'no-errors'

    def is_satisfied(self, output: BenchmarkOutput) -> bool:
        return not (output.not_successful_responses or output.socket_errors)

    def __repr__(self):
        return 'No socket errors and no responses with status not in 2xx or 3xx'


class RequestsPerSecondsGoal(PerformanceGoal):
    """A performance goal satisfied when the number of handled requests
    per second is equals or higher than a given value"""

    type_name = 'requests-per-second'

    def __init__(self, minimum: LimitType):
        self.minimum = float(minimum)

    def is_satisfied(self, output: BenchmarkOutput) -> bool:
        return output.requests_per_second >= self.minimum

    def __repr__(self):
        return f'The minimum amount of handled requests per seconds is {self.minimum}'



import logging
from abc import abstractmethod
from rocore.registry import Registry
from wrktoolbox.results import SuiteReport, BenchmarkOutput


class ReportWriter(Registry):
    """A class that can write a report for a sequence of results."""

    @abstractmethod
    def write(self, report: SuiteReport):
        """Writes a report."""

    @abstractmethod
    def write_output(self, report: SuiteReport, output: BenchmarkOutput):
        """Writes a report output."""


class LogWriter(ReportWriter):
    """A writer that outputs to a logger."""

    type_name = 'log'

    def __init__(self, logger_name: str = 'wrktoolbox'):
        self.logger = logging.getLogger(logger_name)

    def write(self, report: SuiteReport):
        self.logger.info(str(report.suite))

    def write_output(self, report: SuiteReport, output: BenchmarkOutput):
        self.logger.info(str(output))



import json
import pickle
import fnmatch
from base64 import b64decode
from abc import abstractmethod
from pathlib import Path
from typing import Generator, Optional, Sequence
from rocore.exceptions import InvalidArgument
from rocore.typesutils.dateutils import parse_datetime
from wrktoolbox.benchmarks import BenchmarkSuite, PerformanceGoalResult
from wrktoolbox.results import ResultsImporter, SuiteReport, BenchmarkOutput


class FileSystemResultsImporter(ResultsImporter):
    """Base class for importers that can read results from file system"""

    def __init__(self,
                 root_folder: str,
                 filter_urls: Optional[Sequence[str]] = None):
        self._root_path = None
        self.root_path = root_folder
        self._ext_glob_pattern = '*' + self.get_file_extension()
        self.filter_urls = list(filter_urls) if filter_urls else None

    @property
    def root_path(self) -> Path:
        return self._root_path

    @root_path.setter
    def root_path(self, value):
        root_path = Path(value)

        if not root_path.exists():
            raise InvalidArgument('given root path does not exist')

        if not root_path.is_dir():
            raise InvalidArgument('given root path is not a directory')

        self._root_path = root_path

    @abstractmethod
    def parse_suite(self, data: str) -> SuiteReport:
        """Parses a suite"""

    @abstractmethod
    def parse_output(self, data: str) -> BenchmarkOutput:
        """Parses a benchmark output"""

    @abstractmethod
    def get_file_extension(self) -> str:
        """Returns the handled files extension"""

    def _load_suite(self, item: Path) -> SuiteReport:
        with open(str(item), mode='rt', encoding='utf8') as file:
            return self.parse_suite(file.read())

    def _load_output(self, item: Path) -> BenchmarkOutput:
        with open(str(item), mode='rt', encoding='utf8') as file:
            return self.parse_output(file.read())

    def _suites_from_dir(self, folder_path: Path) -> Generator[SuiteReport, None, None]:
        for item in folder_path.iterdir():
            if item.is_symlink():
                continue

            if item.is_dir():
                yield from self._suites_from_dir(item)

            if 'suite' in item.name:
                yield self._load_suite(item)
        return

    def _should_import(self, item: BenchmarkOutput) -> bool:
        if self.filter_urls:
            return any(fnmatch.fnmatch(item.url, pattern) for pattern in self.filter_urls)
        return True

    def _results_from_dir(self, folder_path: Path, report: SuiteReport) -> Generator[BenchmarkOutput, None, None]:
        benchmarks_ids = report.suite.benchmarks_ids

        for item in folder_path.iterdir():
            if item.is_symlink():
                continue

            if item.is_file():
                file_name = str(item)
                if fnmatch.fnmatch(file_name, self._ext_glob_pattern) \
                        and any(benchmark_id in file_name for benchmark_id in benchmarks_ids):
                    result = self._load_output(item)
                    if self._should_import(result):
                        yield result
            else:
                yield from self._results_from_dir(item, report)

    def import_suites(self) -> Generator[SuiteReport, None, None]:
        yield from self._suites_from_dir(self.root_path)

    def import_results(self, report: SuiteReport) -> Generator[BenchmarkOutput, None, None]:
        yield from self._results_from_dir(self.root_path, report)


class BinResultsImporter(FileSystemResultsImporter):

    type_name = 'bin'

    def get_file_extension(self) -> str:
        return '.bin'

    def parse_suite(self, data: str) -> SuiteReport:
        suite = pickle.loads(b64decode(data))  # type: BenchmarkSuite
        return SuiteReport(suite)

    def parse_output(self, data: str) -> BenchmarkOutput:
        output = pickle.loads(b64decode(data))  # type: BenchmarkOutput
        return output


class JsonResultsImporter(FileSystemResultsImporter):

    type_name = 'json'

    def get_file_extension(self) -> str:
        return '.json'

    def parse_suite(self, data: str) -> SuiteReport:
        suite = json.loads(data)
        return SuiteReport(BenchmarkSuite.from_dict(suite))

    def parse_output(self, data: str) -> BenchmarkOutput:
        data = json.loads(data)
        output = BenchmarkOutput.parse('\n'.join(data.get('raw_output')),
                                       data.get('id'),
                                       data.get('suite_id'),
                                       parse_datetime(data.get('start_time')),
                                       parse_datetime(data.get('end_time')))
        output.__dict__['goals_results'] = [PerformanceGoalResult(**item) for item in data.get('goals_results')]
        return output



import click
import logging
from wrktoolbox import version
from wrktoolbox.commands.run import run_command
from wrktoolbox.commands.reports import reports_command
from wrktoolbox.logs import get_app_logger
from .web import disable_ssl_verification


@click.group()
@click.option('--verbose',
              default=False,
              help='Whether to display debug output.',
              is_flag=True)
@click.option('--no-ssl-verify',
              default=False,
              help='Allows to skip ssl verification.',
              is_flag=True)
@click.version_option(version=version)
def main(verbose, no_ssl_verify):
    """
    wrktoolbox is a tool to run HTTP benchmarks with wrk and wrk2 tools, store their output, and generate
    reports.

    For examples and more information, refer to the project in GitHub:
    https://github.com/RobertoPrevato/wrktoolbox
    """
    logger = get_app_logger()
    if verbose:
        logger.setLevel(logging.DEBUG)

    logger.debug('Running in --verbose mode')

    if no_ssl_verify:
        logger.debug('Disabling SSL verification')
        disable_ssl_verification()


main.add_command(run_command)
main.add_command(reports_command)



import os
import uuid
import pickle
from abc import abstractmethod
from base64 import b64encode
from datetime import datetime
from rocore.json import dumps
from rocore.folders import ensure_folder
from wrktoolbox.benchmarks import BenchmarkOutputStore, BenchmarkOutput, BenchmarkConfig, BenchmarkSuite


class FileSystemBenchmarkOutputStore(BenchmarkOutputStore):
    """Base class for file system stores."""

    def __init__(self, output_folder: str = 'out'):
        if output_folder == '$newid':
            output_folder = str(uuid.uuid4())

        ensure_folder(output_folder)
        self.output_folder = output_folder

    def get_file_name(self, prefix: str, suffix: str = '') -> str:
        ts = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        if (prefix == suffix):
            return os.path.join(self.output_folder, f'{prefix}-{ts}') + self.get_file_extension()
        else:
            return os.path.join(self.output_folder, f'{prefix}-{ts}-{suffix}') + self.get_file_extension()

    @abstractmethod
    def get_file_extension(self) -> str:
        """Returns the output files extension"""

    @abstractmethod
    def write_output(self, config: BenchmarkConfig, output: BenchmarkOutput) -> str:
        """Writes output to a str"""

    @abstractmethod
    def write_suite(self, suite: BenchmarkSuite):
        """Writes a suite to a string representation"""

    def store(self, config: BenchmarkConfig, output: BenchmarkOutput):
        with open(self.get_file_name(config.test_id, output.id), mode='wt', encoding='utf8') as output_file:
            output_file.write(self.write_output(config, output))

    def store_suite(self, suite: BenchmarkSuite):
        with open(self.get_file_name('suite', suite.id), mode='wt', encoding='utf8') as output_file:
            output_file.write(self.write_suite(suite))

    def to_dict(self):
        return {
            'type': self.get_class_name(),
            'output_folder': self.output_folder
        }


class BinFileSystemBenchmarkOutputStore(FileSystemBenchmarkOutputStore):
    """A file system store that saves data in binary (pickled) format."""
    type_name = 'bin'

    def get_file_extension(self) -> str:
        return '.bin'

    def write_output(self, config: BenchmarkConfig, output: BenchmarkOutput) -> str:
        return b64encode(pickle.dumps(output)).decode('utf8')

    def write_suite(self, suite: BenchmarkSuite):
        suite.plugins = [plugin.name for plugin in suite.plugins]
        return b64encode(pickle.dumps(suite)).decode('utf8')


class JsonFileSystemBenchmarkOutputStore(FileSystemBenchmarkOutputStore):
    """A file system store that saves data in JSON format."""
    type_name = 'json'

    def get_file_extension(self) -> str:
        return '.json'

    def write_output(self, config: BenchmarkConfig, output: BenchmarkOutput) -> str:
        return dumps(output, indent=4)

    def write_suite(self, suite: BenchmarkSuite):
        return dumps(suite, indent=4)



from uuid import uuid4
from datetime import datetime
from typing import Optional, Union
from pyparsing import Literal, Word, nums, alphanums, OneOrMore, Group, Suppress


unit_chars = 'ums'
decimal_chars = nums + '.'
bytes_size_chars = 'kKMmGgbB'
bytes_size_chars_full = decimal_chars + 'kKMmGgbB'

# Latency   196.94ms  183.71ms 944.41ms   89.18%
latency_pattern = Literal('Latency').suppress() \
  + Word(decimal_chars).setResultsName('latency') \
  + Word(unit_chars).setResultsName('latency_unit') \
  + Word(decimal_chars).setResultsName('stdev') \
  + Word(unit_chars).setResultsName('stdev_unit') \
  + Word(decimal_chars).setResultsName('max_value') \
  + Word(unit_chars).setResultsName('max_unit') \
  + Word(decimal_chars).setResultsName('stdev_perc') \
  + Literal('%')

# Running 5s test @ https://foo.org
head_pattern = Literal("Running").suppress() \
  + Word(nums).setResultsName('duration') \
  + Literal('s').setResultsName('duration_unit') \
  + Literal("test @ ").suppress() \
  + Word(alphanums + "/-.:?&=%").setResultsName('url')

# 10 threads and 10 connections
threads_connections_pattern = Word(nums).setResultsName('threads_count') + Literal('threads and').suppress() \
  + Word(nums).setResultsName('connections_count') + Literal('connections').suppress()

# Req/Sec     7.65      2.98    10.00     71.19%
req_sec_pattern = Literal('Req/Sec').suppress() \
  + Word(decimal_chars).setResultsName('req_sec') \
  + Word(decimal_chars).setResultsName('req_sec_stdev') \
  + Word(decimal_chars).setResultsName('req_sec_max') \
  + Word(decimal_chars).setResultsName('req_sec_stdev_perc') + Literal('%')

# 302 requests in 5.07s, 148.32KB read
# 4294 requests in 30.09s, 2.06MB read
reqs_count_pattern = Word(nums).setResultsName('reqs_count') \
                     + Literal('requests').suppress() \
                     + Literal('in').suppress() \
  + Word(decimal_chars).setResultsName('seconds_count') + Literal('s,').suppress() \
  + Word(decimal_chars).setResultsName('total_transfer_read') \
  + Word(bytes_size_chars).setResultsName('total_transfer_read_unit') \
  + Literal('read').suppress()

# Requests/sec:     59.61
reqs_summary_pattern = Literal('Requests/sec:').suppress() \
  + Word(decimal_chars).setResultsName('reqs_per_second_summary')

# Transfer/sec:     29.28KB
transfer_summary_pattern = Literal('Transfer/sec:').suppress() \
  + Word(decimal_chars).setResultsName('transfer_per_second_summary') \
  + Word(bytes_size_chars).setResultsName('transfer_per_second_summary_unit')

# Latency Distribution
latency_statistics_pattern = Literal('Latency Distribution').suppress() \
  + OneOrMore(Group(Word(nums).setResultsName('percentile')
                    + Literal('%').suppress()
                    + Word(decimal_chars).setResultsName('value')
                    + Word(unit_chars).setResultsName('value_unit'))).setResultsName('values')

"""
  Latency Distribution (HdrHistogram - Recorded Latency)
 50.000%  129.15ms
 75.000%  142.46ms
 90.000%  148.09ms
 99.000%  873.98ms
 99.900%  876.54ms
 99.990%  876.54ms
 99.999%  876.54ms
100.000%  876.54ms

"""
hdrhistogram_pattern = Literal('Latency Distribution (HdrHistogram - Recorded Latency)').suppress() \
    + OneOrMore(Group(Word(decimal_chars) + Suppress('%')
                      + Word(decimal_chars).setResultsName('value')
                      + Word(unit_chars).setResultsName('value_unit'))).setResultsName('values')

"""
#[Mean    =      161.908, StdDeviation   =      150.488]
#[Max     =      876.032, Total count    =           80]
#[Buckets =           27, SubBuckets     =         2048]
"""
HASH, LSB, EQUALS, RSB, COMMA = map(Suppress, '#[=],')
decimal_or_nan = decimal_chars + '-nan' + 'inf'
detailed_percentile_spectrum_pattern = Literal('Detailed Percentile spectrum:').suppress() \
    + Literal('Value').suppress() + Literal('Percentile').suppress() \
    + Literal('TotalCount').suppress() + Literal('1/(1-Percentile)').suppress() \
    + OneOrMore(Group(Word(decimal_or_nan).setResultsName('value') +
                      Word(decimal_or_nan).setResultsName('percentile') +
                      Word(nums).setResultsName('total_count') +
                      Word(decimal_or_nan).setResultsName('percentile_1_1'))).setResultsName('values') \
    + HASH + LSB + Suppress('Mean') + EQUALS + Word(decimal_or_nan).setResultsName('mean') \
    + COMMA + Suppress('StdDeviation') \
    + EQUALS + Word(decimal_or_nan).setResultsName('standard_deviation') + RSB \
    + HASH + LSB + Suppress('Max') + EQUALS + Word(decimal_or_nan).setResultsName('max_value') \
    + COMMA + Suppress('Total count') \
    + EQUALS + Word(nums).setResultsName('total_count') + RSB \
    + HASH + LSB + Suppress('Buckets') + EQUALS + Word(decimal_or_nan).setResultsName('buckets') \
    + COMMA + Suppress('SubBuckets') \
    + EQUALS + Word(nums).setResultsName('sub_buckets') + RSB


comma = Literal(',').suppress()

socket_errors_pattern = Literal('Socket errors: connect').suppress() + Word(nums).setResultsName('connect_errors') \
  + comma \
  + Literal('read').suppress() + Word(nums).setResultsName('read_errors') + comma  \
  + Literal('write').suppress() + Word(nums).setResultsName('write_errors') + comma \
  + Literal('timeout').suppress() + Word(nums).setResultsName('timeout_errors') \

# Non-2xx or 3xx responses: 2400
not_successful_responses_pattern = Literal('Non-2xx or 3xx responses:').suppress() \
                                   + Word(nums).setResultsName('non_2xx_or_3xx_responses_count')


start_pattern = head_pattern + threads_connections_pattern


class ParseFailure:

    def __init__(self, exception_message, desired_type, raw_value):
        self.exception_message = exception_message
        self.desired_type = desired_type
        self.raw_value = raw_value

    def to_dict(self):
        return self.__dict__.copy()


class Result:

    def __setattr__(self, key, value):
        if key in self.__dict__:
            raise AttributeError(f'{key} is read only')
        super().__setattr__(key, value)

    def __repr__(self):
        return ' '.join(f'{key}: {value}' for key, value in self.__dict__.items())

    def __eq__(self, other):
        if other.__class__ is not self.__class__:
            return NotImplemented
        return other.__dict__ == self.__dict__

    @classmethod
    def parse(cls, raw: str):
        try:
            values = cls.pattern.parseString(raw)
        except Exception as pex:
            return ParseFailure(str(pex), cls, raw)
        return cls(**values)

    def to_dict(self):
        return self.__dict__.copy()


class ValueResult(Result):

    def __init__(self, value, unit):
        self.value = value
        self.unit = unit.lower()

    def __repr__(self):
        return f'{self.value}{self.unit}'

    def __eq__(self, other):
        if isinstance(other, ValueResult):
            return other.value == self.value and other.unit == self.unit
        return NotImplemented


class TimeResult(ValueResult):

    def __init__(self, value, unit):
        super().__init__(value, unit)
        self.ms = self._to_ms()

    def _to_ms(self):
        if self.unit == 'ms':
            return self.value
        if self.unit == 'us':
            return self.value / 1000
        if self.unit == 's':
            return self.value * 1000


class SocketErrorsResult(Result):

    pattern = socket_errors_pattern

    def __init__(self,
                 connect_errors,
                 read_errors,
                 write_errors,
                 timeout_errors):
        self.connect_errors = int(connect_errors)
        self.read_errors = int(read_errors)
        self.write_errors = int(write_errors)
        self.timeout_errors = int(timeout_errors)


class LatencyResult(Result):

    pattern = latency_pattern

    def __init__(self,
                 latency,
                 latency_unit,
                 stdev,
                 stdev_unit,
                 max_value,
                 max_unit,
                 stdev_perc):
        self.avg = TimeResult(float(latency), latency_unit)
        self.stdev = TimeResult(float(stdev), stdev_unit)
        self.max = TimeResult(float(max_value), max_unit)
        self.stdev_perc = float(stdev_perc)


class LatencyDistributionResult(Result):
    """wrk latency distribution output"""

    pattern = latency_statistics_pattern

    def __init__(self, values):
        percentiles = {}
        for percentile, value, value_unit in values:
            percentiles[float(percentile)] = TimeResult(float(value), value_unit)
        self.percentiles = percentiles

    def __eq__(self, other):
        if isinstance(other, LatencyDistributionResult):
            return self.percentiles == other.percentiles
        if isinstance(other, dict):
            return self.percentiles == other
        return NotImplemented

    @staticmethod
    def line_matches(value: str):
        return 'Latency Distribution' in value and 'HdrHistogram' not in value

    @staticmethod
    def last_line_matches(value: str):
        return '99% ' in value


class HdrHistogramLatencyDistributionResult(LatencyDistributionResult):
    """wrk2 latency distribution output"""

    pattern = hdrhistogram_pattern

    @staticmethod
    def line_matches(value: str):
        return 'Latency Distribution' in value and 'HdrHistogram' in value

    @staticmethod
    def last_line_matches(value: str):
        return '100.000% ' in value


class RequestsSummaryResult(Result):

    pattern = reqs_summary_pattern

    def __init__(self, reqs_per_second_summary):
        self.reqs_per_second_summary = float(reqs_per_second_summary)


class TransferSummaryResult(Result):

    pattern = transfer_summary_pattern

    def __init__(self, transfer_per_second_summary, transfer_per_second_summary_unit):
        self.transfer_per_second_avg = ValueResult(float(transfer_per_second_summary), transfer_per_second_summary_unit)


class RequestsPerSecondResult(Result):

    pattern = req_sec_pattern

    def __init__(self, req_sec, req_sec_stdev, req_sec_max, req_sec_stdev_perc):
        self.avg = float(req_sec)
        self.stdev = float(req_sec_stdev)
        self.max = float(req_sec_max)
        self.stdev_perc = float(req_sec_stdev_perc)


class TotalRequestsResult(Result):

    pattern = reqs_count_pattern

    def __init__(self, reqs_count, seconds_count, total_transfer_read, total_transfer_read_unit):
        self.requests = int(reqs_count)
        self.seconds = float(seconds_count)
        self.read = ValueResult(float(total_transfer_read), total_transfer_read_unit)


class NotSuccessfulResponses(Result):

    pattern = not_successful_responses_pattern

    def __init__(self, non_2xx_or_3xx_responses_count):
        self.non_2xx_or_3xx_responses_count = int(non_2xx_or_3xx_responses_count)


_non_numeric = {'inf', '-inf', 'nan', '-nan', 'nanus', '-nanus'}  # output from wrk


def try_parse(value, num_type):
    if value in _non_numeric:
        return value
    try:
        return num_type(value)
    except ValueError:
        return value


class DetailedPercentileSpectrumValue(Result):

    def __init__(self, value, percentile, total_count, percentile_1_1):
        self.value = try_parse(value, float)
        self.percentile = try_parse(percentile, float)
        self.total_count = try_parse(total_count, int)
        self.percentile_1_1 = try_parse(percentile_1_1, float)


class DetailedPercentileSpectrum(Result):

    pattern = detailed_percentile_spectrum_pattern

    def __init__(self, values, mean, standard_deviation, max_value, total_count, buckets, sub_buckets):
        self.mean = try_parse(mean, float)
        self.standard_deviation = try_parse(standard_deviation, float)
        self.max = try_parse(max_value, float)
        self.total_count = try_parse(total_count, int)
        self.buckets = try_parse(buckets, int)
        self.sub_buckets = try_parse(sub_buckets, int)
        self.values = [DetailedPercentileSpectrumValue(*value) for value in values]

    @staticmethod
    def line_matches(value: str):
        return 'Detailed Percentile spectrum' in value

    @staticmethod
    def last_line_matches(value: str):
        return '#[Buckets' in value


def all_subclasses(_type):
    yield _type
    for sub_type in _type.__subclasses__():
        yield from all_subclasses(sub_type)


def get_lines(raw_output: str):
    lines = {}
    matching_open_lines = False
    matched_lines = []

    for line in raw_output.splitlines():

        if matching_open_lines:
            matched_lines.append(line)

            if result_type.last_line_matches(line):
                lines[result_type] = '\n'.join(matched_lines)
                matched_lines.clear()
                matching_open_lines = False
            continue

        for result_type in all_subclasses(Result):

            if hasattr(result_type, 'line_matches'):
                if not hasattr(result_type, 'last_line_matches'):
                    raise RuntimeError(f'The class {result_type.__name__} implements a `line_matches` method '
                                       f'but no `last_line_matches` property. Define both methods.')
                if result_type.line_matches(line):
                    matched_lines.append(line)
                    matching_open_lines = True
                    break

            elif hasattr(result_type, 'pattern') and result_type.pattern.matches(line):
                # single line
                lines[result_type] = line
    return lines


def _parse_latency_distribution(line_match):
    if LatencyDistributionResult in line_match:
        return LatencyDistributionResult.parse(line_match[LatencyDistributionResult])

    if HdrHistogramLatencyDistributionResult in line_match:
        return HdrHistogramLatencyDistributionResult.parse(line_match[HdrHistogramLatencyDistributionResult])

    return None


LatencyDistributionType = Union[LatencyDistributionResult, HdrHistogramLatencyDistributionResult, None]


class BenchmarkOutput(Result):

    def __init__(self,
                 *,
                 benchmark_id: str = None,
                 raw_output: str = None,
                 url: str = None,
                 threads: Optional[int] = None,
                 connections: Optional[int] = None,
                 latency: Optional[LatencyResult] = None,
                 duration: Optional[ValueResult] = None,
                 socket_errors: Optional[SocketErrorsResult] = None,
                 detailed_percentile_spectrum: Optional[DetailedPercentileSpectrum] = None,
                 not_successful_responses: int = 0,
                 latency_distribution: LatencyDistributionType = None,
                 requests_summary: Optional[RequestsSummaryResult] = None,
                 requests_per_second: Optional[float] = None,
                 transfer_per_second: Optional[float] = None,
                 total: Optional[TotalRequestsResult] = None,
                 suite_id: Optional[str] = None,
                 start_time: Optional[datetime] = None,
                 end_time: Optional[datetime] = None):
        self.id = benchmark_id or str(uuid4())
        self.raw_output = raw_output
        self.url = url
        self.threads = threads
        self.connections = connections
        self.duration = duration
        self.latency = latency
        self.latency_distribution = latency_distribution
        self.requests_summary = requests_summary,
        self.requests_per_second = requests_per_second
        self.socket_errors = socket_errors
        self.detailed_percentile_spectrum = detailed_percentile_spectrum
        self.not_successful_responses = not_successful_responses
        self.has_errors = bool(socket_errors or not_successful_responses)
        self.transfer_per_second = transfer_per_second
        self.total = total
        self.goals_results = []
        self.suite_id = suite_id
        self.start_time = start_time
        self.end_time = end_time

    def __repr__(self):
        return f'<BenchmarkOutput {self.id} {self.url}>'

    @classmethod
    def parse(cls,
              raw_output: str,
              benchmark_id: Optional[str] = None,
              suite_id: Optional[str] = None,
              start_time: Optional[datetime] = None,
              end_time: Optional[datetime] = None):
        if not benchmark_id:
            benchmark_id = str(uuid4())

        raw_output = raw_output.strip()
        head = start_pattern.parseString(raw_output)

        line_match = get_lines(raw_output)

        duration = TimeResult(int(head.duration), head.duration_unit)
        url = head.url
        threads = int(head.threads_count)
        connections = int(head.connections_count)
        latency = LatencyResult.parse(line_match[LatencyResult]) \
            if LatencyResult in line_match else None
        latency_distribution = _parse_latency_distribution(line_match)

        socket_errors = SocketErrorsResult.parse(line_match[SocketErrorsResult])\
            if SocketErrorsResult in line_match else None

        requests_summary = RequestsPerSecondResult.parse(line_match[RequestsPerSecondResult]) \
            if RequestsPerSecondResult in line_match else None

        requests_per_second = RequestsSummaryResult.parse(line_match[RequestsSummaryResult])\
            .reqs_per_second_summary \
            if RequestsSummaryResult in line_match else None

        transfer_per_second = TransferSummaryResult.parse(line_match[TransferSummaryResult])\
            .transfer_per_second_avg \
            if TransferSummaryResult in line_match else None

        total = TotalRequestsResult.parse(line_match[TotalRequestsResult]) \
            if TotalRequestsResult in line_match else None

        not_successful_responses = NotSuccessfulResponses.parse(line_match[NotSuccessfulResponses])\
            .non_2xx_or_3xx_responses_count \
            if NotSuccessfulResponses in line_match else 0

        detailed_percentile_spectrum = DetailedPercentileSpectrum.parse(line_match[DetailedPercentileSpectrum])\
            if DetailedPercentileSpectrum in line_match else None

        return cls(benchmark_id=benchmark_id,
                   raw_output=raw_output,
                   url=url,
                   threads=threads,
                   connections=connections,
                   latency=latency,
                   duration=duration,
                   socket_errors=socket_errors,
                   not_successful_responses=not_successful_responses,
                   detailed_percentile_spectrum=detailed_percentile_spectrum,
                   latency_distribution=latency_distribution,
                   requests_summary=requests_summary,
                   requests_per_second=requests_per_second,
                   transfer_per_second=transfer_per_second,
                   total=total,
                   suite_id=suite_id,
                   start_time=start_time,
                   end_time=end_time)

    def to_dict(self):
        data = super().to_dict()
        data['raw_output'] = data['raw_output'].splitlines()
        return data




import re
import os
import socket
import urllib.request
from logging import Logger
from wrktoolbox.web import get_ssl_context
from wrktoolbox.benchmarks import BenchmarkSuite
from rocore.decorators import retry


def capture_ips(text: str):
    return re.findall(r'[0-9]+(?:\.[0-9]+){3}', text) + re.findall(r'\b(?:(?:[0-9a-zA-Z]{1,4}|:):?){8}\b', text)


def is_valid_ipv4_address(value):
    try:
        socket.inet_aton(value)
    except socket.error:
        return False
    return True


def is_valid_ipv6_address(value):
    try:
        socket.inet_pton(socket.AF_INET6, value)
    except socket.error:
        return False
    return True


def is_valid_ip_address(value):
    return is_valid_ipv4_address(value) or is_valid_ipv6_address(value)


class IpAddressNotFoundInResponseBody(Exception):

    def __init__(self, source_url, response_body):
        super().__init__('A valid IP address could not be extracted from response body, using configured source url.')
        self.source_url = source_url
        self.response_body = response_body


class InvalidIpAddress(Exception):

    def __init__(self, value):
        super().__init__('The value obtained from configured source url, does not represent a valid IP address.')
        self.value = value


def get_public_ip_address(source_url):
    response = urllib.request.urlopen(source_url, context=get_ssl_context())
    text = response.read().decode('utf8')

    captured_ips = capture_ips(text)

    if not captured_ips:
        raise IpAddressNotFoundInResponseBody(source_url, text)

    value = captured_ips[0]

    if not is_valid_ip_address(value):
        raise InvalidIpAddress(value)

    return value


def setup(suite: BenchmarkSuite, logger: Logger):
    if suite.metadata is None:
        suite.metadata = {}

    source_url = suite.metadata.get('client_ip_source_url',
                                    os.environ.get('WRKTOOLBOX_CLIENT_IP_SOURCE_URL', 'https://api.ipify.org'))

    logger.info('Obtaining client ip address, using source url: %s', source_url)

    def log_retry(exc, attempt):
        if attempt > 3:
            return
        logger.exception('Attempt to obtain public ip address failed; %s', attempt, exc_info=exc)

    @retry(delay=0.5, on_exception=log_retry)
    def get_ip():
        return get_public_ip_address(source_url)

    # noinspection PyBroadException
    try:
        client_ip = get_ip()
    except Exception:
        logger.info('Failed to obtain public ip address')
        suite.public_ip = '???'
    else:
        logger.info('Public client ip: %s', client_ip)
        suite.public_ip = client_ip



import os
import json
import yaml
import urllib.request
from enum import Enum
from typing import Tuple, Mapping
from abc import ABC, abstractmethod
from wrktoolbox.logs import get_app_logger
from rocore.decorators import retry
from rocore.exceptions import InvalidArgument, EmptyArgumentException
from roconfiguration import Configuration, ConfigurationError
from wrktoolbox.web import get_ssl_context


logger = get_app_logger()


class SettingsFormat(Enum):
    JSON = 'json'
    YAML = 'yaml'


class SettingsFileNotFound(InvalidArgument):

    def __init__(self, file_name):
        super().__init__(f'`{file_name}` not found')


def log_retry(exc, attempt):
    if attempt > 3:
        return
    logger.exception('Attempt to downlad settings failed; %s', attempt, exc_info=exc)


class SettingsSource(ABC):

    @abstractmethod
    def is_match(self, value: str) -> bool:
        """Returns a value indicating whether a settings source is handled by this class."""

    @abstractmethod
    def handle(self, value: str) -> Tuple[Mapping, SettingsFormat]:
        """Obtains settings mapping and format from a value."""

    @staticmethod
    def parse(value: str, settings_format: SettingsFormat) -> Mapping:
        # NB: this function violates SRP
        # better code would define a class for each format (low priority here since we support only YAML and JSON)
        if settings_format == SettingsFormat.JSON:
            return json.loads(value)
        if settings_format == SettingsFormat.YAML:
            return yaml.safe_load(value)

        raise RuntimeError(f'Settings format not handled {settings_format}')


class SettingsHttpSource(SettingsSource):

    def is_match(self, value: str) -> bool:
        return value.startswith('http://') or value.startswith('https://')

    @retry(delay=0.5, on_exception=log_retry)
    def _fetch_response(self, source_url):
        return urllib.request.urlopen(source_url, context=get_ssl_context())

    def handle(self, value: str) -> Mapping:
        response = self._fetch_response(value)
        content_type = response.headers['content-type']

        if 'json' in content_type:
            settings_format = SettingsFormat.JSON
        elif 'yaml' in content_type:
            settings_format = SettingsFormat.YAML
        elif '.json' in value:
            settings_format = SettingsFormat.JSON
        elif '.yaml' in value:
            settings_format = SettingsFormat.YAML
        else:
            raise ConfigurationError('Settings must be yaml or json')

        content = response.read().decode('utf8')
        return self.parse(content, settings_format)


class SettingsFileSource(SettingsSource):

    def is_match(self, value: str) -> bool:
        return value.endswith('.json') or value.endswith('.yaml')

    def handle(self, value: str) -> Mapping:
        if not os.path.exists(value):
            raise SettingsFileNotFound(value)

        if not os.path.isfile(value):
            raise InvalidArgument('Invalid settings path: not a file')

        settings_format = SettingsFormat.YAML if value.endswith('.yaml') else SettingsFormat.JSON

        with open(value, mode='rt', encoding='utf8') as settings_file:
            content = settings_file.read()
        return self.parse(content, settings_format)


def normalize_settings(settings_file: str) -> Mapping:
    if not settings_file:
        raise EmptyArgumentException('settings_name')

    for source in [SettingsHttpSource(), SettingsFileSource()]:
        if source.is_match(settings_file):
            return source.handle(settings_file)

    raise InvalidArgument(f'Settings argument `{settings_file}` is not handled.')


def get_configuration(settings_file: str) -> Configuration:
    settings = normalize_settings(settings_file)

    configuration = Configuration(settings)
    configuration.add_environmental_variables('WRKTOOLBOX_', strip_prefix=True)
    return configuration



from setuptools import setup
from wrktoolbox import version


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='wrk-load-tools',
      version=version,
      description='A tool to run wrk/wrk2 benchmarks and store their output.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Operating System :: Unix'
      ],
      url='https://github.com/kishorekumar-kk/wrktoolbox',
      author = 'Kishore Kumar',
      author_email = 'mailtokishorekumar.s@gmail.com',
      keywords='wrk runner benchmarks load performance tests',
      license='MIT',
      packages=['wrktoolbox',
                'wrktoolbox.stores',
                'wrktoolbox.plugins',
                'wrktoolbox.goals',
                'wrktoolbox.commands',
                'wrktoolbox.reports',
                'wrktoolbox.results',
                'wrktoolbox.results.importers'],
      install_requires=['pyparsing',
                        'rocore',
                        'roconfiguration',
                        'click',
                        'certifi'],
      include_package_data=True,
      zip_safe=False,
      entry_points="""
      [console_scripts]
      wrktoolbox=wrktoolbox.main:main
      """)



import os
import yaml
import time
import importlib
import subprocess
import multiprocessing
from collections.abc import Mapping
from uuid import uuid4
from enum import Enum
from logging import Logger
from functools import wraps
from abc import abstractmethod
from typing import Optional, Dict, Sequence, Any
from rocore.exceptions import InvalidArgument, EmptyArgumentException
from rocore.registry import Registry
from rocore.models import Model, String, UInt, Enum as EnumType, Boolean, OfType, Collection, Guid, DateTime
from .wrkoutput import BenchmarkOutput, Result, ParseFailure
from datetime import datetime


class BenchmarkException(Exception):
    """Base class for exceptions happening during benchmarks."""


class ProcessBenchmarkException(BenchmarkException):

    def __init__(self, output: str, exit_code: int):
        super().__init__(f'Process exit code does not indicate success; '
                         f'exit code: {exit_code};'
                         f' output: {output}')


class MissingDependencyException(BenchmarkException):

    def __init__(self):
        super().__init__('Process exit code (127) indicated failure due to missing `wrk` or `wrk2` command; '
                         'please install the necessary dependencies and make sure they are accessible from scripts.')


class BenchmarkPluginException(BenchmarkException):

    def __init__(self, details: str):
        super().__init__(f'Invalid `plugins` configuration; {details}')


class WrkVariant(Enum):
    WRK = 'wrk'
    WRK2 = 'wrk2'


class PerformanceGoalResult(Result):

    def __init__(self,
                 success: bool,
                 goal: str,
                 error: Optional[str] = None):
        self.success = success
        self.goal = goal
        self.error = error

    def to_dict(self):
        data = super().to_dict()
        if not self.error:
            del data['error']
        return data


class PerformanceGoal(Registry):

    @abstractmethod
    def is_satisfied(self, output: BenchmarkOutput) -> bool:
        """Returns a value indicating whether a goal is satisfied."""

    def assert_parsed(self, value: Any):
        assert value is not None
        assert not isinstance(value, ParseFailure)

    @classmethod
    def from_dict(cls, data):
        if 'repr' in data:
            del data['repr']
        return super().from_dict(data)

    def to_dict(self):
        data = self.__dict__.copy()
        data['repr'] = repr(self)
        data['type'] = self.get_class_name()
        return data


class GoalException(BenchmarkException):
    """Base class for performance goal exceptions."""


class BenchmarkConfig(Model):

    test_id = String()
    url = String(nullable=False)
    threads = UInt(nullable=False)
    concurrency = UInt(nullable=False)
    duration = UInt(nullable=False)
    timeout = UInt(nullable=False)
    script = String()
    app_variant = EnumType(WrkVariant)
    responses_per_second = UInt()
    headers = OfType(dict)
    latency_statistics = Boolean()
    repeat = UInt()
    goals = Collection(PerformanceGoal)

    def __init__(self,
                 url: str,
                 threads: int = -1,
                 concurrency: int = 10,
                 duration: int = 20,
                 timeout: int = 20,
                 script: str = None,
                 app_variant: WrkVariant = WrkVariant.WRK,
                 responses_per_second: Optional[int] = None,
                 headers: Optional[Dict[str, str]] = None,
                 latency_statistics: Optional[bool] = True,
                 test_id: str = None,
                 repeat: int = 1,
                 goals: Optional[Sequence[PerformanceGoal]] = None):
        if threads < 1 or threads is None:
            threads = multiprocessing.cpu_count()

        if app_variant == WrkVariant.WRK2 and responses_per_second is None:
            responses_per_second = 10

        if responses_per_second is not None and responses_per_second > 0:
            app_variant = WrkVariant.WRK2

        self.test_id = test_id or uuid4()
        self.url = url
        self.threads = threads
        self.concurrency = concurrency
        self.duration = duration
        self.timeout = timeout
        self.script = script
        self.app_variant = WrkVariant(app_variant or 'wrk')
        self.responses_per_second = responses_per_second
        self.latency_statistics = bool(latency_statistics)
        self.headers = headers
        self.repeat = repeat
        self.goals = goals

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.url}>'

    def _get_responses_per_second(self):
        return f' -R{self.responses_per_second}' \
            if self.responses_per_second and self.app_variant == WrkVariant.WRK2 else ''

    def _get_headers(self):
        if not self.headers:
            return ''
        parts = []
        for key, value in self.headers.items():
            parts.append(f'-H "{key}: {value}"')
        return ' ' + ' '.join(parts)

    def __eq__(self, other):
        if isinstance(other, BenchmarkConfig):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def to_dict(self):
        return {
            'test_id': self.test_id,
            'url': self.url,
            'threads': self.threads,
            'concurrency': self.concurrency,
            'duration': self.duration,
            'timeout': self.timeout,
            'script': self.script,
            'app_variant': self.app_variant.value,
            'responses_per_second': self.responses_per_second,
            'latency_statistics': self.latency_statistics,
            'headers': self.headers
        }

    def get_cmd(self):
        return f'{self.app_variant.value} {self.url} ' \
               f'-c {self.concurrency} ' \
               f'-t {self.threads} ' \
               f'-d {self.duration} ' \
               f'--timeout {self.timeout}' \
               + (' --latency' if self.latency_statistics else '') \
               + (f' -s {self.script}' if self.script else '') \
               + self._get_responses_per_second() \
               + self._get_headers()


class Benchmark(Model):

    id = Guid()
    config = OfType(BenchmarkConfig)

    def __init__(self, config: BenchmarkConfig):
        self.id = uuid4()
        self.config = config

    def run(self, logger=None, suite_id=None) -> BenchmarkOutput:
        config = self.config
        start_time = datetime.utcnow()

        p = subprocess.Popen(config.get_cmd(), shell=True, stdout=subprocess.PIPE)
        p.wait(config.duration + 12)

        end_time = datetime.utcnow()

        output = p.stdout.read().decode()

        if logger and output:
            logger.debug(f'[*] Process output: \n\n{output}\n\n')

        if p.returncode != 0:
            # something went wrong when running the benchmark
            if p.returncode == 127:
                raise MissingDependencyException()

            raise ProcessBenchmarkException(output, p.returncode)

        return BenchmarkOutput.parse(output,
                                     benchmark_id=config.test_id,
                                     suite_id=suite_id,
                                     start_time=start_time,
                                     end_time=end_time)


class BenchmarkOutputStore(Registry):

    @abstractmethod
    def store(self, config: BenchmarkConfig, output: BenchmarkOutput):
        """Stores the results of a benchmark."""

    @abstractmethod
    def store_suite(self, suite: 'BenchmarkSuite'):
        """Stores information about a suite."""


def exception_handle(catch_exc, exc_type):
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except catch_exc as e:
                raise exc_type(str(e))
        return wrapper
    return decorator


class BenchmarkPlugin:

    def __init__(self, module):
        if isinstance(module, Mapping):
            module = module.get('module')
        if isinstance(module, str):
            module = importlib.import_module(module)
        self.module = module

    @property
    def setup(self):
        return getattr(self.module, 'setup')

    @property
    def has_setup(self) -> bool:
        return hasattr(self.module, 'setup')

    @property
    def name(self):
        return self.module.__name__

    def to_dict(self):
        return {'module': self.name,
                'has_setup': self.has_setup}

    def __repr__(self):
        return f'<Plugin {self.name}>'


@exception_handle(ModuleNotFoundError, BenchmarkPluginException)
def handle_plugins(data):
    plugins = data.get('plugins')
    if not plugins:
        return

    for plugin in plugins:
        yield BenchmarkPlugin(plugin)


def _get_goals(array):
    if not array:
        return []
    return [PerformanceGoal.from_configuration(item) for item in array]


class HostData:

    def __init__(self,
                 cpu_count: Optional[int] = None,
                 env: Optional[Mapping] = None):
        if cpu_count is None:
            cpu_count = multiprocessing.cpu_count()
        if env is None:
            env = os.environ.copy()
        self.cpu_count = cpu_count
        self.env = env

    def to_dict(self):
        return self.__dict__.copy()


class BenchmarkSuite(Model):
    id = String(nullable=False)
    public_ip = String()
    location = String()
    scripts_folder = String()
    configurations = Collection(BenchmarkConfig)
    stores = Collection(BenchmarkOutputStore)
    plugins = Collection((str, BenchmarkPlugin))
    goals = Collection(PerformanceGoal)
    benchmarks_ids = Collection(str)
    start_time = DateTime()
    end_time = DateTime()
    think_time = UInt(nullable=False)

    root_settings = {'threads',
                     'concurrency',
                     'duration',
                     'timeout',
                     'responses_per_second',
                     'headers',
                     'latency_statistics',
                     'repeat'}

    def __init__(self,
                 configurations: Sequence[BenchmarkConfig],
                 stores: Sequence[BenchmarkOutputStore],
                 scripts_folder: str,
                 plugins: Optional[Sequence[BenchmarkPlugin]] = None,
                 goals: Optional[Sequence[PerformanceGoal]] = None,
                 think_time: int = 0,
                 _id: Optional[str] = None,
                 benchmarks_ids: Optional[Sequence[str]] = None,
                 public_ip: Optional[str] = None,
                 metadata: Optional[Any] = None,
                 start_time: Optional[datetime] = None,
                 end_time: Optional[datetime] = None,
                 host_data: Optional[HostData] = None):
        if host_data is None:
            host_data = HostData()
        if benchmarks_ids is None:
            benchmarks_ids = []
        if think_time is None:
            think_time = 0
        self.id = _id or uuid4()
        self.stores = stores
        self.scripts_folder = scripts_folder
        self.configurations = configurations
        self.plugins = plugins
        self.goals = goals
        self.think_time = think_time
        self.host = host_data
        self.metadata = metadata
        self.public_ip = public_ip
        self.location = self._get_location_from_env(host_data)
        self.start_time = start_time
        self.end_time = end_time
        self.benchmarks_ids = benchmarks_ids
        self._check_configurations_ids(configurations)

        if scripts_folder:
            for configuration in configurations:
                if configuration.script:
                    configuration.script = os.path.join(scripts_folder, configuration.script)

    def load_plugins(self):
        if not self.plugins:
            return
        if any(isinstance(plugin, str) for plugin in self.plugins):
            self.plugins = [BenchmarkPlugin(name) for name in self.plugins]

    def estimated_time(self) -> int:
        """Returns an estimated time required for completion, in seconds.
        This estimate does not count time spent by implementations of output store."""
        i = 0
        for configuration in self.configurations:
            i += configuration.duration * configuration.repeat

            if configuration.app_variant == WrkVariant.WRK2:
                # thread calibration may take about 10 seconds
                i += 10

            i += self.think_time
        # one second margin added, since code execution time is not counted
        return i + 1 - self.think_time

    @staticmethod
    def _check_configurations_ids(configurations: Sequence[BenchmarkConfig]):
        found_ids = set()

        for i, configuration in enumerate(configurations):
            if configuration.test_id in found_ids:
                raise InvalidArgument(f'Duplicated test id: {configuration.test_id}; '
                                      f'change test_id configuration for your benchmarks.')
            if not configuration.test_id:
                configuration.test_id = f'test_{i}'

            found_ids.add(configuration.test_id)

    def __repr__(self):
        return f'<{self.__class__.__name__} {len(self)}>'

    def __str__(self):
        return (f'Suite {self.id}; '
                f'Configurations {", ".join(configuration.test_id for configuration in self.configurations)};')

    def __len__(self):
        return len(self.configurations)

    def run(self, logger: Logger):
        logger.info('Estimated time %s s', self.estimated_time())
        self.start_time = datetime.utcnow()
        configurations_count = len(self.configurations)

        for configuration in self.configurations:
            if not configuration.repeat:
                continue

            for i in range(configuration.repeat):
                benchmark = Benchmark(configuration)
                benchmark.suite_id = self.id

                logger.info(f'Running benchmark...\n{configuration.get_cmd()}')

                output = benchmark.run(logger, self.id)

                self.benchmarks_ids.append(output.id)

                self.check_goals(configuration, output, logger)

                logger.debug(f'Storing output for benchmark {benchmark.id}...')
                self.store_output(configuration, output)

                logger.info('---')

                if self.think_time and i + 1 < configurations_count:
                    logger.debug(f'Waiting for {self.think_time} seconds')
                    time.sleep(self.think_time)

        logger.debug(f'Storing suite data')
        self.end_time = datetime.utcnow()
        self.store_self()

    def check_goals(self, configuration: BenchmarkConfig, output: BenchmarkOutput, logger: Logger):
        if not self.goals and not configuration.goals:
            logger.debug(f'No performance goals are defined for {configuration.test_id}')
        if self.goals:
            self._check_goals(output, self.goals, logger)
        if configuration.goals:
            self._check_goals(output, configuration.goals, logger)

    @staticmethod
    def _check_goals(output: BenchmarkOutput,
                     goals: Sequence[PerformanceGoal],
                     logger: Logger):
        for goal in goals:
            try:
                is_satisfied = goal.is_satisfied(output)
            except (AssertionError, GoalException) as error:
                logger.exception('Error while checking performance goal', exc_info=error)
                result = PerformanceGoalResult(False, repr(goal), str(error))
            else:
                logger.debug(f'--> goal {goal.get_class_name()} satisfied: {is_satisfied}')
                result = PerformanceGoalResult(is_satisfied, repr(goal), None)

            output.goals_results.append(result)

    def store_self(self):
        for store in self.stores:
            store.store_suite(self)

    def store_output(self,
                     configuration: BenchmarkConfig,
                     output: BenchmarkOutput):
        for store in self.stores:
            store.store(configuration, output)

    @classmethod
    def from_yaml(cls, file_path: str):
        with open(file_path, mode='rt', encoding='utf8') as settings_file:
            data = yaml.safe_load(settings_file)
        return cls.from_dict(data)

    @staticmethod
    def _get_location_from_env(host: HostData):
        return host.env.get('WRKTOOLBOX_LOCATION', host.env.get('LOCATION', ''))

    def to_dict(self):
        return {
            'id': self.id,
            'scripts_folder': self.scripts_folder,
            'configurations': self.configurations,
            'stores': self.stores,
            'plugins': self.plugins,
            'goals': self.goals,
            'think_time': self.think_time,
            'metadata': self.metadata,
            'host': self.host,
            'public_ip': self.public_ip,
            'benchmarks_ids': self.benchmarks_ids,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'location': self.location
        }

    @staticmethod
    def use_base_url(data):
        if not data or 'configurations' not in data:
            raise EmptyArgumentException('mapping')

        base_url = data.get('base_url')
        base_url_key = data.get('base_url_key', '$BASEURL')

        if not base_url:
            return

        configurations = data.get('configurations')
        for configuration in configurations:
            url = configuration.get('url')  # type: str

            if not url:
                configuration['url'] = base_url
                continue

            if base_url_key in url:
                configuration['url'] = url.replace(base_url_key, base_url)
                continue

            if not url.startswith('http'):
                configuration['url'] = base_url + url

        metadata = data.get('metadata')
        if not metadata:
            return

        for key, value in metadata.items():
            if base_url_key in value:
                metadata[key] = value.replace(base_url_key, base_url)

    @staticmethod
    def use_root_settings(data):
        if not data or 'configurations' not in data:
            raise EmptyArgumentException('mapping')

        configurations = data.get('configurations')

        for root_name in BenchmarkSuite.root_settings:
            root_value = data.get(root_name)

            if not root_value:
                continue

            for configuration in configurations:
                if root_name not in configuration or configuration.get(root_name) is None:
                    configuration[root_name] = root_value

    @staticmethod
    def normalize_configuration(data):
        configurations_key = 'configurations'
        benchmarks_key = 'benchmarks'
        stores_key = 'stores'

        # support calling configurations 'benchmarks' for user friendliness in configuration files
        if configurations_key not in data:
            if benchmarks_key in data:
                data[configurations_key] = data[benchmarks_key]
                del data[benchmarks_key]
            else:
                raise InvalidArgument(f'Missing benchmarks configuration: either specify '
                                      f'it with `{configurations_key}` or `{benchmarks_key}`')

        for name in {configurations_key, stores_key}:
            if name not in data:
                raise InvalidArgument(f'Missing `{name}` in configuration')

        BenchmarkSuite.use_base_url(data)
        BenchmarkSuite.use_root_settings(data)

    @classmethod
    def from_dict(cls, data):
        cls.normalize_configuration(data)

        # NB: plugins must be first imported, as they might register new types
        plugins = list(handle_plugins(data))

        for benchmark in data.get('configurations'):
            if 'goals' in benchmark:
                benchmark['goals'] = _get_goals(benchmark.get('goals'))

        host_data = HostData(**data.get('host')) if 'host' in data else None

        return cls([BenchmarkConfig(**item) for item in data.get('configurations')],
                   [BenchmarkOutputStore.from_configuration(item) for item in data.get('stores')],
                   data.get('scripts_folder'),
                   plugins,
                   _get_goals(data.get('goals')),
                   data.get('think_time'),
                   _id=data.get('id'),
                   benchmarks_ids=data.get('benchmarks_ids'),
                   public_ip=data.get('public_ip'),
                   metadata=data.get('metadata'),
                   start_time=data.get('start_time'),
                   end_time=data.get('end_time'),
                   host_data=host_data)



import ssl
import certifi

SECURE_SSLCONTEXT = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=certifi.where())
SECURE_SSLCONTEXT.check_hostname = True

INSECURE_SSLCONTEXT = ssl.SSLContext()
INSECURE_SSLCONTEXT.check_hostname = False


ssl_verify = True


def get_ssl_context() -> ssl.SSLContext:
    global ssl_verify

    if ssl_verify:
        return SECURE_SSLCONTEXT

    return INSECURE_SSLCONTEXT


def disable_ssl_verification():
    global ssl_verify
    ssl_verify = False




import reprlib
from typing import Union
from wrktoolbox.benchmarks import PerformanceGoal, GoalException
from wrktoolbox.wrkoutput import BenchmarkOutput
from .common import LimitType


PercentileType = Union[int, float, str]


class AverageLatencyGoal(PerformanceGoal):
    """A performance goal that is satisfied when the overall average latency
    is lower than a given limit, in milliseconds"""

    type_name = 'avg-latency'

    def __init__(self, limit: LimitType):
        """
        Creates a new instance of AverageLatencyGoal with given limit in ms.

        :param limit: average latency limit in milliseconds
        """
        self.limit = float(limit)

    def __repr__(self):
        return f'Average latency for web requests must be less than {self.limit} ms.'

    def is_satisfied(self, output: BenchmarkOutput) -> bool:
        self.assert_parsed(output.latency)
        return output.latency.avg.ms <= self.limit


class PercentileLatencyGoal(PerformanceGoal):
    """A performance goal that is satisfied when the percentile latency of web requests is less than a limit in ms."""

    type_name = 'percentile-latency'

    def __init__(self, percentile: PercentileType, limit: LimitType):
        """
        Creates a new instance of PercentileLatencyGoal with given percentile and limit in ms.

        :param percentile: reference percentile of this goal
        :param limit: average latency limit in milliseconds
        """
        self.percentile = float(percentile)
        self.limit = limit

    def __repr__(self):
        return f'The {self.percentile} percentile latency of web requests must be less than {self.limit} ms.'

    def is_satisfied(self, output: BenchmarkOutput) -> bool:
        self.assert_parsed(output.latency_distribution)
        value = output.latency_distribution.percentiles.get(self.percentile)

        if value is None:
            raise GoalException(f'Percentile {self.percentile} is not found among output percentiles, '
                                f'therefore it`s not possible to use an exact value. '
                                f'Configure performance goals to use percentiles returned by wrk. '
                                f'Found percentiles are: {reprlib.repr(output.latency_distribution.percentiles)}')

        return value.ms <= self.limit

